from cumin.objectselector import ObjectSelector, ObjectLinkColumn, ObjectTable, \
    ObjectTableColumn, ObjectQmfTable, ObjectQmfSelectorTable, ObjectSelectorTaskForm, ObjectSelectorTask
from cumin.objectframe import ObjectFrame, ObjectTaskForm
from cumin.util import *
from cumin.qmfadapter import ObjectQmfAdapter
from cumin.task import Task, TaskLink
from cumin.objectframe import ObjectDetails

from wooly.forms import StringField, ScalarField, OptionInputSet

from hadoop import *
from hadoopobjectframe import *

from sage.util import *

from wooly.util import StringCatalog

strings = StringCatalog(__file__)

class DataNodeAdapter(HadoopAdapter):
    def get_sage_results(self, values):
        owner_filter = self.get_owner_filter(values['session'])
        results = self.app.remote.get_data_node_list(owner=owner_filter)
        return results

class DataNodeSelector(ObjectSelector):
    def __init__(self, app, name, module, check_owner=False):
        cls = app.model.com_redhat_cumin_grid_hadoop.DataNode

        super(DataNodeSelector, self).__init__(app, name, cls)

        self.module = module

        self.add_search_filter(self.table.id_col)
        self.table.adapter = DataNodeAdapter(app, cls, check_owner)

        task = DataNodeCreate(app, name + ".create", self.module)
        link = TaskLink(app, "dataNode_create", task)
        self.links.add_child(link)

        DataNodeDelete(app, self, "dataNode_delete", self.module)

        self.enable_csv_export()

    def create_table(self, app, name, cls):
        return DataNodeTable(app, name, cls)

    def render_title(self, session):
        return "HDFS data nodes"

class DataNodeTable(ObjectQmfSelectorTable):
    def __init__(self, app, name, cls):
        super(DataNodeTable, self).__init__(app, name, cls)

        self.id_col = self.DataNodeColumn(app, "idcol", cls.Id, cls.Id, "datanodeframe")
        self.state_col = ObjectTableColumn(app, "statecol", cls.State)
        self.uptime_col = UptimeColumn(app, "uptimecol", cls.Uptime)
        self.owner_col = ObjectTableColumn(app, "ownercol", cls.Owner)
        # we need to be able to filter by parent_id, but don't need to see it
        self.parent_id_col = ObjectTableColumn(app, "parentid", cls.Parent_id)
        self.parent_id_col.visible = False

        self.add_column(self.id_col)
        self.add_column(self.state_col)
        self.add_column(self.uptime_col)
        self.add_column(self.owner_col)
        self.add_column(self.parent_id_col)

    def get_data_options(self, session):
        options = super(DataNodeTable, self).get_data_options(session)
        options.parent_filter = {"column":self.parent_id_col, "value":self.frame.id.get(session)}
        return options

    class DataNodeColumn(ObjectLinkColumn):
        def render_cell_href(self, session, record):
            id = unescape_entity(record[self.id_field.index])
            frame = self.frame.children_by_name[self.frame_path]
            return frame.get_href(session, id, frame.parent.host.get(session))

class DataNodeDelete(HadoopNodeDeleteTask):
    def __init__(self, app, selector, name, module):
        super(DataNodeDelete, self).__init__(app, selector)
        cls = app.model.com_redhat_cumin_grid_hadoop.DataNode

        self.module = module

        self.form = DataNodeDeleteForm(app, self.name, self, cls)
        self.invoc = None

    def do_invoke(self, invoc, ids, args):
        self.invoc = invoc
        (hadoophost) = args
        self.app.remote.stop_data_node(hadoophost, ids, invoc.make_callback())

    def get_title(self, session):
        return "Remove data nodes"

    def get_item_content(self, session, item):
        return xml_escape(item)


class DataNodeDeleteForm(HadoopNodeDeleteForm):
    def process_submit(self, session):
        self.validate(session)
        ids = self.selection.get(session)
        url = self.return_url.get(session)
        hadoophost = self.hadoophost.get(session)
        self.page.redirect.set(session, url)

        if not self.errors.get(session):
            self.task.invoke(session, ids, (hadoophost))


    def render_title(self, session):
        return "Remove data nodes"

class DataNodeCreate(Task):
    def __init__(self, app, name, module):
        super(DataNodeCreate, self).__init__(app, name)
        cls = app.model.com_redhat_cumin_grid_hadoop.DataNode

        self.module = module

        self.form = DataNodeCreateForm(app, self.name + ".createform", self, cls)
        self.invoc = None

    def callback(self, result):
        if result == False:
            self.invoc.status = self.invoc.FAILED
        self.invoc.end()

    def do_invoke(self, session, object, invoc, args):
        self.invoc = invoc
        (binfile, owner, hadoophost, description, count, name_node) = args

        try:
            self.invoc = invoc
            self.app.remote.start_data_node(hadoophost, name_node, binfile, owner, description, count, invoc.make_callback())
        except Exception, e:
            invoc.status = invoc.FAILED
            log.debug("Creating data node failed", exc_info=True)
            invoc.end()

    def get_title(self, session, x):
        return "Create data nodes"

    def enter(self, session, obj):
        form_session = wooly.Session(self.app.form_page)

        if obj:
            self.form.id.set(form_session, obj._id)
        hadoop_schedulers = self.app.remote.get_hosts("SCHEDULER", "HADOOP")
        self.form.hadoophost.set(form_session, hadoop_schedulers[0])
        self.form.nn_id.set(form_session, self.module.frame.children_by_name[self.module.frame.name + ".namenodeframe"].id.get(session))
        self.form.return_url.set(form_session, session.marshal())
        self.form.show(form_session)

        self.do_enter(session, obj, form_session)
        return form_session

    def get_href(self, session, obj):

        return self.enter(session, obj).marshal()


class DataNodeCreateForm(HadoopNodeCreateForm):
    def __init__(self, app, name, task, cls):
        super(DataNodeCreateForm, self).__init__(app, name, task, cls)

        self.binfile = BinFileLoc(app, name + ".binfile")
        self.add_field(self.binfile)

        self.count = self.CountField(app, name + ".count")
        self.add_field(self.count)

        self.hadoophost = Parameter(app, "hadoophost")
        self.add_parameter(self.hadoophost)

        self.nn_id = Parameter(app, "nn_id")
        self.add_parameter(self.nn_id)


    def process_display(self, session):
        self.scheduler.validate(session)

    def process_submit(self, session):
        self.validate(session)

        url = self.return_url.get(session)
        self.page.redirect.set(session, url)

        if not self.errors.get(session):
            binfile = self.binfile.get(session)
            owner = session.client_session.username()
            hadoophost = self.hadoophost.get(session)
            count = self.count.get(session)
            name_node = self.nn_id.get(session)
            description = ""
            self.task.invoke(session, None, (binfile, owner, hadoophost, description, count, name_node))
            self.task.exit_with_redirect(session, url)

    def render_title(self, session):
        return "Create data nodes"

    class CountField(IntegerField):
        def __init__(self, app, name):
            super(DataNodeCreateForm.CountField, self).__init__(app, name)
            self.input.param.default = 1
            self.input.size = 4

        def render_title(self, session):
            return "Number of data nodes to start"

class DataNodeFrame(HadoopObjectFrame):
    def __init__(self, app, name, module):
        super(DataNodeFrame, self).__init__(app, name, module)

        self.view = DataNodeObjectView(app, "view", self.object)
        self.replace_child(self.view)

        self.details = DataNodeDetails(self.app, self.name + ".details", self.object)
        self.view.add_tab(self.details)

    def get_title(self, session):
        retval = "Data node:'%s' (%s)" % (xml_escape(self.object.get(session).Name), xml_escape(self.object.get(session).Id))
        return retval

    def get_object(self, session, id, ipc):
        host = self.frame.host.get(session)
        response = self.app.remote.get_data_node(self.frame.host.get(session), [id])
        obj = HadoopObject(response, self.app.model.com_redhat_cumin_grid_hadoop.DataNode)
        return obj

class DataNodeObjectView(HadoopObjectView):
    def add_details_tab(self):
        pass

    def render_title(self, session):
        obj = self.object.get(session)
        if obj is not None:
            return "Data node: %s" % xml_escape(obj.Id)
        else:
            return ""

class DataNodeDetails(HadoopObjectDetails):
    def __init__(self, app, name, object):
        super(DataNodeDetails, self).__init__(app, name, object)
        self.object = object

    def render_title(self, session):
        return "Details"

    def render_name(self, session):
        '''sample....would only be called if DataNodeDetails display was overridden to say {name}'''
        obj = self.object.get(session)
        return obj.Name
