from work import Work
from util import now_minus_duration
from datetime import datetime, timedelta

from mint.update import AviaryCollectorUpdate, AviarySubmitterUpdate
import suds.sax.text
import logging

log = logging.getLogger("mint.aviary.collector")

class BaseCollectorObj(object):
    def __init__(self, endpoint_id, res_id, class_name):

        self.pkg_name = "com.redhat.grid"
        self.class_name = class_name
        self.endpoint_id = endpoint_id
        self.id = res_id

        # For most things we want a Pool attr which doesn't
        # exist in the classads....
        try:
            self.Pool = str(self.id.pool)
        except:
            pass

    def set_from_ad(self, attrs):
        for attr in attrs:
            if attr.value == "UNDEFINED":
                continue
            if type(attr.value) == suds.sax.text.Text:
                v = str(attr.value)
            else:
                v = attr.value
            setattr(self, attr.name, v)

    def set_from_summary(self, id, summaries):
        pass

class BaseCollectorQuery(Work):
    def __init__(self, ep):
        super(BaseCollectorQuery, self).__init__()
        self.ep = ep

        self.names = []
        self.strip_host = False
        self.objects = dict()

        self.recycle_interval = timedelta(seconds=10)
        self.max_silence = timedelta(seconds=30)
        self.no_summary = True
        self.no_classads = False

        self.deleted = False

    def wrap_update(self, aviary_obj):
        return AviaryCollectorUpdate(self.ep.app.model, aviary_obj)

    def make_update(self, id, attrs, summaries):
        aviary_obj = self.make_collector_obj(id)
        aviary_obj.set_from_summary(id, summaries)
        aviary_obj.set_from_ad(attrs)
        return self.wrap_update(aviary_obj)

    def make_delete(self, id):
        aviary_obj = self.make_collector_obj(id)
        up = AviaryCollectorUpdate(self.ep.app.model, aviary_obj)
        up.delete = True
        return up

    def make_collector_obj(self, id):
        raise Exception("NotImplemented")

    def get_objects(self, no_summary):
        raise Exception("NotImplemented")

    def get_attributes(self, ids):
        ads = []
        final = False
        stat = False
        chunk_size = 3

        chunks = [ids[i:i+chunk_size] for i in range(0, len(ids), chunk_size)]

        for chunk in chunks:
            stat, res = self.ep.app.aviary_ops.get_attributes(self.ep.url,
                                                              chunk,
                                                              self.names,
                                                              self.strip_host)
            if stat == 0:
                final = 0
                ads.extend(res)

        return final, ads
        
    def do_the_job(self, foreman):

        # If the endpoint has been abandoned, then we should delete the objects
        # that we've created and mark ourselves deleted.  This will also prevent requeueing
        if self.ep.abandoned:
            for name, id in self.objects.iteritems():
                update = self.make_delete(id[0])
                self.ep.app.update_thread.enqueue(update)

            self.objects = []
            self.deleted = True
            return

        # Note that if we cannot talk to this endpoint because the endpoint has gone
        # away, the locator thread will eventually mark it abandoned if the
        # locator is in use
        log.debug("Gathering values")
        stat, objs = self.get_objects(self.no_summary)
        if stat == 0:
            summaries = dict()
            if self.no_summary:
                # We're going to get everything from classads, we
                # did not ask for summaries
                ids = [o.id for o in objs \
                       if not hasattr(o, "status_code") or o.status.code == "OK"]
            else:
                # We'll get things from classads and/or summaries.
                # Save the summaries in a dictionary so we can look them up
                ids = []
                for o in objs:
                    if o.status.code == "OK":
                        ids.append(o.id)
                        if hasattr(o.id, "address"):
                            k = (o.id.name, o.id.address)
                        else:
                            k = (o.id.name, "")
                        summaries[k] = o.summary

            if self.no_classads:
                # all we have are ids and summaries
                log.debug("Queueing updates, summary based")
                for i in ids:
                    update = self.make_update(i, [], summaries)
                    if update:
                        self.ep.app.update_thread.enqueue(update)
            else:
                # get the classad in addition to the ids and summaries
                log.debug("Gathering attributes")
                stat, ads = self.get_attributes(ids)
                if stat == 0:
                    log.debug("Queueing updates, attribute based")
                    for ad in ads:
                        if ad.status.code == "OK":
                            update = self.make_update(ad.id, 
                                                      ad.ad.attrs, summaries)
                            if update:
                                self.ep.app.update_thread.enqueue(update)
                
            # Remember the objects that we have just seen
            for id in ids:
                # HA names end with @
                if id.name.endswith("@"):
                    search = id.name
                else:
                    search = id.name
                    if hasattr(id, "address"):
                        search += "/"+ id.address
                if search in self.objects:
                    self.objects[search][1] = datetime.now()
                else:
                    self.objects[search] = [id, datetime.now()]

        # Look for objects we haven't heard from in a while
        deleted = []
        now = datetime.now()
        for name, id in self.objects.iteritems():
            if now - id[1] >= self.max_silence:
                # We need to remove this object
                update = self.make_delete(id[0])
                self.ep.app.update_thread.enqueue(update)
                deleted.append(name)
        for d in deleted:
            del self.objects[d]
        del deleted
            
    def completed(self):
        return self.ep.abandoned and self.deleted

class CollectorObj(BaseCollectorObj):
    def __init__(self, endpoint_id, res_id):
        super(CollectorObj, self).__init__(endpoint_id, res_id, "Collector")

    def set_from_ad(self, attrs):
        super(CollectorObj, self).set_from_ad(attrs)

        # Have to adapt because the cumin db
        # refers to Machine as System
        if hasattr(self, "Machine"):
            setattr(self, "System", self.Machine)

class QueryCollector(BaseCollectorQuery):
    def __init__(self, ep):
        super(QueryCollector, self).__init__(ep)

        self.names = ["CondorPlatform",
                      "CondorVersion",
                      "Machine",
                      "MonitorSelfAge",
                      "MonitorSelfCPUUsage",
                      "MonitorSelfImageSize",
                      "MonitorSelfRegisteredSocketCount",
                      "MonitorSelfResidentSetSize",
                      "MonitorSelfTime",
                      "DaemonStartTime",
                      "Name",
                      "MyAddress",
                      "RunningJobs",
                      "IdleJobs",
                      "HostsTotal",
                      "HostsClaimed",
                      "HostsUnclaimed",
                      "HostsOwner"]

    def make_collector_obj(self, id):
        return CollectorObj(self.ep.id, id)

    def get_objects(self, no_summary):
        return self.ep.app.aviary_ops.get_collector(self.ep.url, no_summary=no_summary)

class NegotiatorObj(BaseCollectorObj):
    def __init__(self, endpoint_id, res_id):
        super(NegotiatorObj, self).__init__(endpoint_id, res_id, "Negotiator")

        self.pref = "LastNegotiationCycle"
        self.suff = "0"

    def set_from_ad(self, attrs):
        for attr in attrs:
            if attr.value == "UNDEFINED":
                continue

            # Strip the extra verbiage from the negotiator cycle attrs
            if type(attr.value) == suds.sax.text.Text:
                v = str(attr.value)
            else:
                v = attr.value
            if attr.name.startswith(self.pref) and attr.name.endswith(self.suff):
                name = attr.name[len(self.pref):-len(self.suff)]
                setattr(self, name, v)
            else:
                setattr(self, attr.name, v)

        self.System = self.Machine

class QueryNegotiator(BaseCollectorQuery):
    def __init__(self, ep):
        super(QueryNegotiator, self).__init__(ep)

        self.names = ["CondorPlatform",
                      "CondorVersion",
                      "Machine",
                      "MonitorSelfAge",
                      "MonitorSelfCPUUsage",
                      "MonitorSelfImageSize",
                      "MonitorSelfRegisteredSocketCount",
                      "MonitorSelfResidentSetSize",
                      "MonitorSelfTime",
                      "DaemonStartTime",
                      "Name",
                      "MyAddress",

                      # the following ones will have to be remapped
                      "LastNegotiationCycleMatchRate0",
                      "LastNegotiationCycleMatches0",
                      "LastNegotiationCycleDuration0",
                      "LastNegotiationCycleMatchRateSustained0",
                      "LastNegotiationCyclePhase1Duration0",
                      "LastNegotiationCyclePhase2Duration0",
                      "LastNegotiationCyclePhase3Duration0",
                      "LastNegotiationCyclePhase4Duration0",
                      "LastNegotiationCycleSlotShareIter0",
                      "LastNegotiationCycleSubmittersFailed0",
                      "LastNegotiationCycleSubmittersOutOfTime0",
                      "LastNegotiationCycleSubmittersShareLimit0",
                      "LastNegotiationCycleTime0",
                      "LastNegotiationCycleEnd0",
                      "LastNegotiationCyclePeriod0",
                      "LastNegotiationCycleNumSchedulers0",
                      "LastNegotiationCycleActiveSubmitterCount0",
                      "LastNegotiationCycleNumIdleJobs0",
                      "LastNegotiationCycleNumJobsConsidered0",
                      "LastNegotiationCycleRejections0",
                      "LastNegotiationCycleTotalSlots0",
                      "LastNegotiationCycleCandidateSlots0",
                      "LastNegotiationCycleTrimmedSlots0"]

    def make_collector_obj(self, id):
        return NegotiatorObj(self.ep.id, id)

    def get_objects(self, no_summary):
        return self.ep.app.aviary_ops.get_negotiator(self.ep.url, no_summary=no_summary)

class SubmitterObj(BaseCollectorObj):
    def __init__(self, endpoint_id, res_id):
        super(SubmitterObj, self).__init__(endpoint_id, res_id, "Submitter")

    def set_from_summary(self, id, summaries):
        self.Machine = str(id.machine)
        self.Name = str(id.name)

        summary = summaries[(id.name,"")]
        self.HeldJobs = summary.held
        self.IdleJobs = summary.idle
        self.RunningJobs = summary.running
        self.ScheddName = str(id.scheduler)
        self.Owner = ""
        self.JobQueueBirthdate = 0

class QuerySubmitter(BaseCollectorQuery):
    def __init__(self, ep):
        super(QuerySubmitter, self).__init__(ep)

        # We're going to get everything from the summary
        # record in this case
        self.no_summary = False
        self.no_classads = True

    def make_collector_obj(self, id):
        return SubmitterObj(self.ep.id, id)

    def get_objects(self, no_summary):
        return self.ep.app.aviary_ops.get_submitter(self.ep.url, no_summary=no_summary)

    def wrap_update(self, aviary_obj):
        return AviarySubmitterUpdate(self.ep.app.model, aviary_obj)

class SchedulerObj(BaseCollectorObj):
    def __init__(self, endpoint_id, res_id):
        super(SchedulerObj, self).__init__(endpoint_id, res_id, "Scheduler")

    def set_from_ad(self, attrs):
        super(SchedulerObj, self).set_from_ad(attrs)

        # Have to adapt because the cumin db
        # refers to Machine as System
        if hasattr(self, "Machine"):
            setattr(self, "System", self.Machine)

        # Special stuff that maybe we can remap
        def remap(name, target):
            if not hasattr(self, name) and hasattr(self, target):
                setattr(self, name, getattr(self, target)) 

        def remap_div(name, operand1, operand2):
            if not hasattr(self, name) and \
               hasattr(self, operand1) and hasattr(self, operand2):

                try:
                    op1 = float(getattr(self, operand1))
                    op2 = float(getattr(self, operand2))
                    if op2 == 0:
                        setattr(self, name, 0)
                    else:
                        setattr(self, name, op1 / op2)
                except:
                    pass

        remap("WindowedStatWidth",   "RecentStatsLifetime")
        remap("ShadowExceptions",    "RecentJobsExitException")
        remap("JobsSubmittedCum",    "JobsSubmitted")
        remap("JobsCompletedCum",    "JobsCompleted")
        remap("JobsExitedCum",       "JobsExited")
        remap("ShadowExceptionsCum", "JobsExitException")
        remap("JobsStartedCum",      "JobsStarted")
        remap("SumRunningTimeCum",   "JobsAccumRunningTime")
        remap("SumTimeToStartCum",   "JobsAccumTimeToStart")

        remap_div("JobSubmissionRate", 
                  "RecentJobsSubmitted", 
                  "RecentStatsLifetime")

        remap_div("JobCompletionRate",
                  "RecentJobsCompleted",
                  "RecentStatsLifetime")

        remap_div("JobStartRate",
                  "RecentJobsStarted",
                  "RecentStatsLifetime")

        remap_div("MeanTimeToStartCum",
                  "JobsAccumTimeToStart",
                  "JobsStarted")

        remap_div("MeanRunningTimeCum",
                  "JobsAccumRunningTime",
                  "JobsCompleted")

        remap_div("MeanTimeToStart",
                  "RecentJobsAccumTimeToStart",
                  "RecentJobsStarted")

        remap_div("MeanRunningTime",
                  "RecentJobsAccumRunningTime",
                  "RecentJobsCompleted")

class QueryScheduler(BaseCollectorQuery):
    def __init__(self, ep):
        super(QueryScheduler, self).__init__(ep)

        # start with all the attributes based on condor-7.6
        self.names = ["CondorPlatform",
                      "CondorVersion",
                      "MonitorSelfAge",
                      "MonitorSelfCPUUsage",
                      "MonitorSelfImageSize",
                      "MonitorSelfRegisteredSocketCount",
                      "MonitorSelfResidentSetSize",
                      "MonitorSelfTime",
                      "DaemonStartTime",

                      "JobQueueBirthdate",
                      "MaxJobsRunning",
                      "Machine",
                      "MyAddress",
                      "NumUsers",
                      "Name",
                      "TotalHeldJobs",
                      "TotalIdleJobs",
                      "TotalJobAds",
                      "TotalRemovedJobs",
                      "TotalRunningJobs",
                      "WindowedStatWidth",
                      "UpdateInterval",
                      "JobsSubmitted",
                      "JobSubmissionRate",
                      "JobsCompleted",
                      "JobCompletionRate",
                      "JobsExited",
                      "ShadowExceptions",
                      "JobsSubmittedCum",
                      "JobsCompletedCum",
                      "JobsExitedCum",
                      "ShadowExceptionsCum",
                      "JobsStartedCum",
                      "JobsStarted",
                      "JobStartRate",
                      "MeanTimeToStartCum",
                      "MeanRunningTimeCum",
                      "SumTimeToStartCum",
                      "SumRunningTimeCum",
                      "MeanTimeToStart",
                      "MeanRunningTime",

        # Get the remap targets, too.  These are new
        # names in 7.8.  If the 7.6 names aren't there,
        # we will remap to these
                      "RecentStatsLifetime",
                      "RecentJobsExitException",
                      "JobsSubmitted",
                      "JobsCompleted",
                      "JobsExited",
                      "JobsExitException",
                      "JobsStarted",
                      "JobsAccumRunningTime",
                      "JobsAccumTimeToStart",
                      "RecentJobsSubmitted", 
                      "RecentStatsLifetime",
                      "RecentJobsCompleted",
                      "RecentStatsLifetime",
                      "RecentJobsStarted",
                      "RecentStatsLifetime",
                      "JobsAccumTimeToStart",
                      "JobsStarted",
                      "JobsAccumRunningTime",
                      "JobsCompleted",
                      "RecentJobsAccumTimeToStart",
                      "RecentJobsStarted",
                      "RecentJobsAccumRunningTime",
                      "RecentJobsCompleted"]

    def make_collector_obj(self, id):
        return SchedulerObj(self.ep.id, id)

    def get_objects(self, no_summary):
        return self.ep.app.aviary_ops.get_scheduler(self.ep.url, no_summary=no_summary)

class SlotObj(BaseCollectorObj):
    def __init__(self, endpoint_id, res_id):
        super(SlotObj, self).__init__(endpoint_id, res_id, "Slot")

    def set_from_ad(self, attrs):
        super(SlotObj, self).set_from_ad(attrs)

        # Have to adapt because the cumin db
        # refers to Machine as System
        if hasattr(self, "Machine"):
            setattr(self, "System", self.Machine)

class QuerySlot(BaseCollectorQuery):
    def __init__(self, ep):
        super(QuerySlot, self).__init__(ep)

        # Initialize the name index for paging
        # We are extending the basic collector endpoint here
        self.ep.idx = None
        self.ep.size = 16

        self.strip_host = False

        self.names = ["CondorPlatform",
                      "CondorVersion",
                      "DaemonStartTime",
                      "MonitorSelfAge",
                      "MonitorSelfCPUUsage",
                      "MonitorSelfImageSize",
                      "MonitorSelfRegisteredSocketCount",
                      "MonitorSelfResidentSetSize",
                      "MonitorSelfTime",
                      "AccountingGroup",
                      "Activity",
                      "Arch",
                      "CheckpointPlatform",
                      "ClientMachine",
                      "ClockDay",
                      "ClockMin",
                      "ConcurrencyLimits",
                      "CondorLoadAvg",
                      "ConsoleIdle",
                      "Cpus",
                      "CurrentRank",
                      "Disk",
                      "EnteredCurrentActivity",
                      "EnteredCurrentState",
                      "FileSystemDomain",
                      "GlobalJobId",
                      "ImageSize",
                      "IsValidCheckpointPlatform",
                      "JobId",
                      "JobStart",
                      "KeyboardIdle",
                      "KFlops",
                      "LastBenchmark",
                      "LastFetchWorkCompleted",
                      "LastFetchWorkSpawned",
                      "LastPeriodicCheckpoint",
                      "LoadAvg",
                      "Machine",
                      "MaxJobRetirementTime",
                      "Memory",
                      "Mips",
                      "MyAddress",
                      "MyCurrentTime",
                      "Name",
                      "NextFetchWorkDelay",
                      "OpSys",
                      "PreemptingConcurrencyLimits",
                      "PreemptingOwner",
                      "PreemptingUser",
                      "PreemptingRank",
                      "RemoteOwner",
                      "RemoteUser",
                      "Requirements",
                      "Rank",
                      "SlotID",
                      "Start",
                      "StarterAbilityList",
                      "State",
                      "TimeToLive",
                      "TotalClaimRunTime",
                      "TotalClaimSuspendTime",
                      "TotalCondorLoadAvg",
                      "TotalCpus",
                      "TotalDisk",
                      "TotalJobRunTime",
                      "TotalJobSuspendTime",
                      "TotalLoadAvg",
                      "TotalMemory",
                      "TotalSlots",
                      "TotalTimeBackfillBusy",
                      "TotalTimeBackfillIdle",
                      "TotalTimeBackfillKilling",
                      "TotalTimeClaimedBusy",
                      "TotalTimeClaimedIdle",
                      "TotalTimeClaimedRetiring",
                      "TotalTimeClaimedSuspended",
                      "TotalTimeMatchedIdle",
                      "TotalTimeOwnerIdle",
                      "TotalTimePreemptingKilling",
                      "TotalTimePreemptingVacating",
                      "TotalTimeUnclaimedBenchmarking",
                      "TotalTimeUnclaimedIdle",
                      "TotalVirtualMemory",
                      "UidDomain",
                      "UpdateInterval",
                      "VirtualMemory",
                      "WindowsBuildNumber",
                      "WindowsMajorVersion",
                      "WindowsMinorVersion"]

    def make_collector_obj(self, id):
        return SlotObj(self.ep.id, id)

    def get_objects(self, no_summary):
        if self.ep.idx is None:
            mode = None
        else:
            mode = "AFTER"

        status, ids = self.ep.app.aviary_ops.get_slot_id(self.ep.url, 
                                                         self.ep.size,
                                                         mode,
                                                         self.ep.idx,
                                                         no_summary=no_summary)

        if ids.remaining == 0:
            self.ep.idx = None
        else:
            self.ep.idx = ids.results[-1].id

        # We can't return the whole thing because the caller expects
        # only a list, not the "remaining"
        return status, ids.results

