from util import now_minus_duration
from datetime import datetime
from submissions import QuerySubmissionsBefore, QuerySubmissionsAfter
from collector import QueryCollector, QueryNegotiator, QuerySubmitter, \
QueryScheduler, QuerySlot
from sysimage import QueryWallarooMeta

class Endpoint(object):
    def __init__(self, url, id):
        self.url = url
        self.id = id
        self.abandoned = False
        
    def abandon(self):
        self.abandoned = True

    def get_work_list(self):
        return None

class SubmissionEndpoint(Endpoint):
    def __init__(self, url, id, factory):
        super(SubmissionEndpoint, self).__init__(url, id)
        self.size = 8    # do we want to control size per endpoint record?
        self.app = factory.app

        my_classes = ("Submission")
        self.classes = [c for c in factory.abbrev_classes \
                            if c in my_classes]

        # Set up time spans for polling/history and initialize search indices
        self.init_query_indices(factory.poll_delta, 
                                factory.history_delta)

    def init_query_indices(self, poll_delta, history_delta):

        now = datetime.now()

        # Probably want to do the "AFTER" queries first.
        # We only do this once.
        self.after_queries_completed = False

        # When the AFTER queries have been completed, 
        # we start the BEFORE queries.  Restart when
        # we reach the history threshold.
        self.before_queries_completed = False

        # We will actively poll submissions with qdates in the range
        # now() - poll_delta to now().  This will be a moving 
        # window, as now() moves into the future and we continuously
        # loop over submissions for polling purposes.  Set a starting point.
        self.poll_delta = poll_delta
        self.after_index = now_minus_duration(self.poll_delta, now)

        # We will get historical data for submissions with qdates in 
        # the range now() - poll_delta to now() - history_delta.
        # We'll loop around to the beginning of time and start polling
        # again just in case, but the latency may be really long.
        self.history_delta = history_delta
        self.before_index = self.after_index
        self.before_cutoff = now_minus_duration(self.history_delta, now)

    def get_work_list(self):
        if "Submission" in self.classes:
            return (QuerySubmissionsAfter(self),
                    QuerySubmissionsBefore(self),)
        return ()
        
class CollectorEndpoint(Endpoint):
    def __init__(self, url, id, factory):
        super(CollectorEndpoint, self).__init__(url, id)        

        self.app = factory.app

        my_classes = ("Collector", "Negotiator", "Submitter", "Scheduler", "Slot")
        self.classes = [c for c in factory.abbrev_classes \
                            if c in my_classes]

    def get_work_list(self):
        work = []
        if "Collector" in self.classes:
            work.append(QueryCollector(self))

        if "Negotiator" in self.classes:
            work.append(QueryNegotiator(self))

        if "Submitter" in self.classes:
            work.append(QuerySubmitter(self))

        if "Scheduler" in self.classes:
            work.append(QueryScheduler(self))

        if "Slot" in self.classes:
            work.append(QuerySlot(self))

        return work

class WallarooEndpoint(Endpoint):
    def __init__(self, url, id, factory):
        super(WallarooEndpoint, self).__init__(url, id)

        self.url = url
        self.app = factory.app

    def get_work_list(self):
        return [QueryWallarooMeta(self)]

class EndpointFactory(object):
    def __init__(self, app, classes, poll_delta, history_delta):
        '''
        classes is a list of (package, class) tuples containing
        the names of packages and classes that should be polled.
        The package and class names come from the Rosemary data
        model and determine which endpoint types we are interested in
        and which classes provided by that endpoint we are interested in.
        '''
        self.app = app

        self.classes = classes
        self.abbrev_classes = [c[1] for c in self.classes \
                                   if c[0] in ("com.redhat.grid",
                                               "com.redhat.sesame")]
        
        self.poll_delta = poll_delta
        self.history_delta = history_delta

        self.type_map = \
          {("CUSTOM","QUERY_SERVER"): self._make_submission_endpoint,
           ("COLLECTOR", "POOL"): self._make_collector_endpoint,
           ("WALLAROO", "META"): self._make_wallaroo_endpoint}
    
    def get_endpoint_types(self):
        # This routine maps Rosemary class names to
        # endpoint types (resource, sub_type) so a locator class can know
        # what types of endpoints to report

        classes_to_endpoints = \
            {"Collector":  ("COLLECTOR", "POOL"),
             "Negotiator": ("COLLECTOR", "POOL"),
             "Scheduler":  ("COLLECTOR", "POOL"),
             "Submitter":  ("COLLECTOR", "POOL"),
             "Slot":       ("COLLECTOR", "POOL"),
             "Submission": ("CUSTOM", "QUERY_SERVER"),
             "Sysimage":   ("WALLAROO", "META")}

        endpoints = set()
        for c in self.abbrev_classes:
            try:
                endpoints.add(classes_to_endpoints[c])
            except:
                pass
        return list(endpoints)

    def _make_collector_endpoint(self, url, id):
        return CollectorEndpoint(url, id, self)

    def _make_submission_endpoint(self, url, id):
        return SubmissionEndpoint(url, id, self)

    def _make_wallaroo_endpoint(self, url=None, id=None):
        if not url:
            url = self.app.wallaroo_host
        return WallarooEndpoint(url, id, self)

    def make_endpoint(self, url, id):
        try:
            return self.type_map[(id.resource, id.sub_type)](url, id)
        except KeyError:
            return Endpoint(url, id)
