import logging

from threading import Thread
from time import sleep
from mint.util import MintBlockingDaemonThread

log = logging.getLogger("mint.aviary")

class LocatorThread(MintBlockingDaemonThread):
    def __init__(self, data_providers, foreman, endpoint_factory):

        # We don't care about app for this thread,
        # at least not currently
        super(LocatorThread, self).__init__(None)

        self.data_providers = data_providers
        self.foreman = foreman
        self.ep_fact = endpoint_factory
        self.daemon = True
        self.stop_requested = False
        self.interval_secs = 30

        self.start()

    def run(self):

        # Find out from the endpoint factory what endpoint
        # types we should query for.  This is a list of
        # (resource, subtype) tuples.
        endpoint_types = self.ep_fact.get_endpoint_types()

        def block():
            self._condition.acquire()
            if not self.stop_requested:
                self._condition.wait(self.interval_secs)
            self._condition.release()

        def gather_endpoints(endpoint_types):
            ends = []
            for provider in self.data_providers:
                ends.extend(provider.get_endpoints_list(endpoint_types))
            return ends

        endpoints = dict()
        while not self.stop_requested:
            try:
                ends = gather_endpoints(endpoint_types)
            except Exception:
                import traceback
                message = traceback.format_exc()
                log.debug("LocatorThread: %s" % message)
                block()
                continue

            locs = dict([[(e.id.name, 
                           e.id.resource,
                           e.id.sub_type), e] for e in ends])
            loc_keys = set(locs.keys())

            # the endpoints that we already know about
            keys = set(endpoints.keys())

            # loc_keys - keys gives the items returned from Aviary that are 
            # not in the current dictionary, ie new stuff
            new_endpoints = loc_keys - keys

            # keys - loc_keys gives the items that are not currently in Aviary, ie
            # old stuff that has been deleted
            stale_endpoints = keys - loc_keys

            for e in stale_endpoints:
                # if a thread is currently processing the dictionary entry,
                # that's okay because with Python reference counting the record
                # will still be valid.  When the thread is done processing, the record
                # will disappear.
                endpoints[e].abandon()
                del endpoints[e]

            for e in new_endpoints:
                endpoints[e] = self.ep_fact.make_endpoint(locs[e].location[0], 
                                                          locs[e].id)
                self.foreman.add_work(endpoints[e].get_work_list())

            block()
