import logging

from work import Foreman
from endpoints import EndpointFactory
from locator import LocatorThread
from datetime import timedelta
from sage.util import SynthEndpoint, ResourceID

log = logging.getLogger("mint.aviary.session")

class MintAviarySession(object):
    def __init__(self, app, classes):
        self.app = app

        self.classes = classes
        self.foreman = None
        self.locatorThread = None

        # need to know the location of the aviary query server(s)
        #    get this from sage?
        #    we don't have an object to start from so we will have to
        #    ask for a list of endpoints to poll, I suppose

        # self.app.aviary_ops has the aviary interface, initialized
        # with the locator, etc.

        # For each class that we are gathering, we need to launch a thread
        # or set of threads that can handle the querying of aviary.  This
        # thread will need to know what interface to call, and what to do
        # with the data it gets back, and how to push it into the update
        # queue.  Hopefully this can be a general set of classes, and maybe
        # we key the specific implementations by the model.class value...

    def start(self):
        log.info("Starting %s", self)
   
        # Handle bind by class
        classes = []
        for cls in self.classes:
            pname = cls._package._name
            cname = cls._name
            classes.append((pname, cname))

            log.info("Querying for Aviary class %s.%s" % (pname, cname))

        self.foreman = Foreman(num_workers=8)

        # This class stores which types of endpoints we want to poll
        # and knows what queries we want to make for each endpoint we find.
        # The locator thread uses this class and the data_providers below
        # to find valid endpoints and generate work for the Foreman.
        endpoint_factory = EndpointFactory(self.app,
                                           classes,
                                           poll_delta=timedelta(days=64),
                                           history_delta=timedelta(weeks=52))

        # These objects return the raw list of known data providers that
        # can be polled for data.  The locator thread will poll them
        # for data providers (contrained by endpoint type)
        data_providers = (self.app.aviary_ops,
                          WallarooEndpoints(self.app.wallaroo_host))

        self.locatorThread = LocatorThread(data_providers, 
                                           self.foreman, 
                                           endpoint_factory)

    def stop(self):
        log.info("Stopping %s", self)

        if self.foreman:
            self.foreman.end_work()

        if self.locatorThread:
            self.locatorThread.stop()
            self.locatorThread.join()
        log.info("Stopped %s", self)

class WallarooEndpoints(object):
    def __init__(self, wallaroo_host):
        self.wallaroo_host = wallaroo_host

    def get_endpoints_list(self, endpoint_types):
        ends = []
        if ("WALLAROO","META") in endpoint_types:
            res_id = ResourceID("WALLAROO", 
                                sub_type="META", 
                                name=self.wallaroo_host)
            url_strs = [self.wallaroo_host]
            ends.append(SynthEndpoint(res_id, 
                                      url_strs))
        return ends
