import bisect
import logging

from threading import Thread, Lock, Condition
from Queue import Queue
from time import sleep
from datetime import timedelta, datetime

log = logging.getLogger("mint.aviary.work")

class Work(object):
    def __init__(self):

        # Set this to None for no delay on reruning
        # the task when it is returned to the queue
        self.recycle_interval = timedelta(seconds=10)

    def do_the_job(self, foreman):
        pass

    def completed(self):
        return True

class Foreman(object):
    def __init__(self, num_workers=1):

        self.work_to_do = Queue()
        self.stop_requested = False

        self.workers = []
        for _ in range(num_workers):
            self.workers.append(Worker(self))

        self.waiting = WaitList(self.work_to_do)

    def end_work(self):
        # Add an empty work object on the queue so that any worker 
        # waiting will be woken up with a "get"
        if not self.stop_requested:
            self.stop_requested = True
            self.waiting.stop()
            for _ in range(len(self.workers)):
                self.work_to_do.put(Work())
            for w in self.workers:
                w.join()

    def add_work(self, work_list):
        if not self.stop_requested and work_list:
            for work in work_list:
                self.work_to_do.put(work)

    def request_work(self, current=None):
        work = None

        # If we have a task that is not yet completed
        # check to see whether the task should be 
        # continued by this thread or requeued
        if current is not None and \
           not current.completed():

            # If there is a recycle interval specified,
            # put the task on the wait list
            if current.recycle_interval is not None:
                self.waiting.add_waiter(current, 
                  datetime.now() + current.recycle_interval)
            elif self.work_to_do.empty():
                work = current
            else:
                self.work_to_do.put(current)

        if work is None:
            work = self.work_to_do.get()
 
        if self.stop_requested:
            # Signal thread to stop
            work = None
        return work

class Worker(Thread):
    def __init__(self, foreman):
        Thread.__init__(self)
        self.foreman = foreman
        self.daemon = True
        self.stop_requested = False
        self.start()
    
    def run(self):
        work = None
        while True:
            work = self.foreman.request_work(work)
            if work is None:
                break # quittin' time
            work.do_the_job(self.foreman)

class WaitList(Thread):
    def __init__(self, ready_queue):
        super(WaitList, self).__init__()

        self.waiting = []
        self.ready_queue = ready_queue

        self._lock = Lock()
        self._condition = Condition(self._lock)

        self.stop_requested = False
        
        self.start()

    def stop(self, timeout=5):
        self._condition.acquire()
        self.stop_requested = True
        self._condition.notify()
        self._condition.release()

        if self.isAlive():
            self.join(timeout)
            log.info("%s stopped" % self.getName())

    def add_waiter(self, work, alarm):
        self._condition.acquire()
        
        # Ask where this will be inserted, and then insert it
        idx = bisect.bisect_right(self.waiting, (alarm, work))
        bisect.insort_right(self.waiting, (alarm, work))

        # If we are inserting the first item on the list we
        # want to wake up the thread so it can reset its sleep time
        if idx == 0:
            self._condition.notify()

        self._condition.release()

    def run(self):
        self._condition.acquire()
        while True:
            # Sleep forever or until the alarm time
            # for the first item on the list
            if self.waiting:
                d = self.waiting[0][0] - datetime.now()
                s = d.days * 86400.0 + d.microseconds / 1000000.0 + d.seconds
            else:
                s = None
            if s is None or s > 0:
                self._condition.wait(s)

            # Might be time to shut down...
            if self.stop_requested:
                break

            # See if something finished...
            if self.waiting and \
               datetime.now() >= self.waiting[0][0]:
                work = self.waiting.pop(0)[1]
                self.ready_queue.put(work)
                work = None
        self._condition.release()
