from database import MintDatabase
from expire import ExpireThread
from model import MintModel
from aviary.session import MintAviarySession
from update import UpdateThread
from vacuum import VacuumThread
from cumin.admin import *
from util import *

log = logging.getLogger("mint.main")

class Mint(object):
    def __init__(self, home, model_paths, database_dsn):
        self.home = home

        model_dir = [os.path.join(self.home, x) for x in model_paths]

        self.model = MintModel(self, model_dir)
        self.model.sql_logging_enabled = False
        self.database = MintDatabase(self, database_dsn)
        self.admin = CuminAdmin(self)
        self.update_thread = UpdateThread(self)
        self.aviary_session = None

        self.expire_enabled = True
        self.expire_thread = ExpireThread(self)

        self.vacuum_enabled = True
        self.vacuum_thread = VacuumThread(self)

        self.print_event_level = 0

        # List of classes to bind in QMF. Referenced by session and update thread.
        self._qmf_classes = set()

        # List of classes to gather via Aviary.  Derived from self._qmf_classes
        # in combination with the list of classes supported by Aviary.
        self._aviary_classes = set()

        # If self._qmf_classes is not set, this is the list of packages
        # to bind.  This will be translated into self.qmf_classes.
        self._qmf_packages = set()

        # Things we know should not be bound.
        self._qmf_package_filter = ["com.redhat.cumin", "com.redhat.cumin.grid"]

        # Aviary interface.
        self.aviary_query_servers = ""
        self.aviary_collector_servers = ""
        self.aviary_key = ""
        self.aviary_cert = ""
        self.aviary_root_cert = ""
        self.aviary_domain_verify=True
        self.aviary_locator = ""

        # Wallaroo interface for inventory data
        self.wallaroo_host = ""
        
        # For development use only
        self.aviary_suds_logs = False
        self.aviary_prefer_condor = True
        self.aviary_ops = None

    def set_classes(self, classes):
        self._qmf_classes = classes

    def get_bound_classes(self):
        return self._aviary_classes.union(self._qmf_classes)

    def check(self):
        log.info("Checking %s", self)

        self.model.check()
        self.database.check()
        self.model.init()

    def _init_aviary(self):
        # If aviary_query_servers or aviary_locator is defined
        # here, then we are going to get some classes from Aviary
        # instead of QMF.  Initialize a MintAviary object and
        # eliminate classes from the set of things bound by
        # the QMF interface.
        if not (self.aviary_query_servers or self.aviary_locator):
            return

        supported_classes = [self.model.com_redhat_grid.Submission,
                             self.model.com_redhat_grid.Collector,
                             self.model.com_redhat_grid.Negotiator,
                             self.model.com_redhat_grid.Submitter,
                             self.model.com_redhat_grid.Scheduler,
                             self.model.com_redhat_grid.Slot,
                             self.model.com_redhat_sesame.Sysimage]

#        supported_classes = [self.model.com_redhat_grid.Submitter,
#                             self.model.com_redhat_grid.Scheduler]

        for c in supported_classes:
            if c in self._qmf_classes:
                self._qmf_classes.remove(c)
                self._aviary_classes.add(c)

        if not self._aviary_classes:
            # hmm, no supported classes, nothing to do
            return

        imports_ok = True
        try:
            from sage.aviary.aviaryoperations import \
                SudsLogging, AviaryOperations
        except:
            imports_ok = False
        if imports_ok:
            SudsLogging.set(self.aviary_suds_logs, self.home)

            # By default Cumin uses /var/lib/condor/aviary/services for
            # aviary wsdl files if it exists.
            aviary_dir = ["/var/lib/condor/aviary/services",
                          os.path.join(self.home, "rpc-defs/aviary")]
            if not self.aviary_prefer_condor:
                aviary_dir = [aviary_dir[1], aviary_dir[0]]

            # Aviary will choose an impl that gives us operations
            # depending on the xxx_servers arguments.
            # If locator is non empty, their actual values will be overridden 
            # but the presence of a value will still control enable/disable for
            # operations on that server type.
            self.aviary_ops = AviaryOperations(
                                      "aviary", aviary_dir,
                                      self.aviary_locator,
                                      None, # we don't care about job_servers
                                      self.aviary_query_servers,
                                      None, # or hadoop servers
                                      self.aviary_collector_servers,
                                      key=self.aviary_key, 
                                      cert=self.aviary_cert,
                                      root_cert=self.aviary_root_cert,
                                      domain_verify=self.aviary_domain_verify)
        else:
            log.info("Imports failed for Aviary interface, disabling")

        # This session type sets up discovery for data objects supported by Aviary
        if self.aviary_ops:
            self.aviary_session = MintAviarySession(self, self._aviary_classes)

        log.info("%s Aviary locator interface" % \
                ((self.aviary_locator and \
                  (self.aviary_query_servers or \
                   self.aviary_collector_servers) and \
                  imports_ok) and "Enabled" or "Disabled"))
                
        log.info("%s Aviary interface for query operations." % \
                 ((self.aviary_query_servers and imports_ok) and "Enabled" or "Disabled"))

        log.info("%s Aviary interface for collector operations." % \
                 ((self.aviary_collector_servers and imports_ok) and "Enabled" or "Disabled"))

    def init(self):
        log.info("Initializing %s", self)

        def state(cond):
            return cond and "enabled" or "disabled"

        log.info("Expiration is %s", state(self.expire_enabled))
        log.info("Vacuum is %s", state(self.vacuum_enabled))

        # If self._qmf_classes has not been set, then build classes
        # here from the package list.  We want to be canonical, ie
        # always have things expressed in terms of classes
        if not self._qmf_classes:
            if not self._qmf_packages:
                self._qmf_packages = self.model._packages
            for pkg in self._qmf_packages:
                if pkg._name not in self._qmf_package_filter:
                    self._qmf_classes = self._qmf_classes.union(set(pkg._classes))
        else:            
            # Apply the package filter to the specified class list
            black_list = set()
            for cls in self._qmf_classes:
                if cls._package._name in self._qmf_package_filter:
                    black_list.add(cls)
            self._qmf_classes.difference_update(black_list)

        # Set up Aviary interface for getting some classes...
        # (someday it will be ALL classes)
        self._init_aviary()

        self.database.init()

        self.update_thread.init()
        self.expire_thread.init()
        self.vacuum_thread.init()

    def start(self):
        log.info("Starting %s", self)
        
        self.update_thread.start()

        if self.aviary_session:
            self.aviary_session.start()

        if self.expire_enabled:
            self.expire_thread.start()

        if self.vacuum_enabled:
            self.vacuum_thread.start()

    def stop(self, skip_vacuum=False):
        log.info("Stopping %s", self)

        self.update_thread.stop()

        if self.expire_enabled:
            self.expire_thread.stop()
        
        if self.vacuum_enabled and not skip_vacuum:
            self.vacuum_thread.stop()

        if self.aviary_session:
            self.aviary_session.stop()
        log.info("Session stopped")

    def __repr__(self):
        return self.__class__.__name__
