import os

from mint.database import MintDatabase
from mint.expire import ExpireThread
from mint.model import MintModel
from mint.vacuum import VacuumThread
from session import PlumageSession
from update import UpdateThread
from cumin.admin import CuminAdmin
#from util import *

import logging

log = logging.getLogger("mint.plumage")

class Plumage(object):
    def __init__(self, home, model_paths, 
                 server_host, server_port, database_dsn):

        model_dir = [os.path.join(home, x) for x in model_paths]

        # Can we just use a different model dir here?
        # That would control visible packages, vacuuming, etc.
        self.model = MintModel(self, model_dir)
        self.model.sql_logging_enabled = False

        self.database = MintDatabase(self, database_dsn)
        self.admin = CuminAdmin(self)
        self.update_thread = UpdateThread(self)
        self.session = PlumageSession(self, server_host, server_port)

        self.expire_enabled = True
        self.expire_thread = ExpireThread(self)

        self.vacuum_enabled = True
        self.vacuum_thread = VacuumThread(self)
        
        self.plumage_host = "localhost"
        self.plumage_port = 27017

        self.print_event_level = 0

        self._packages = set()

        self._classes = set()

        self._package_filter = ["com.redhat.cumin"]

    def set_classes(self, classes):
        self._classes = classes

    def get_bound_classes(self):
        return self._classes

    def check(self):
        log.info("Checking %s", self)

        self.model.check()
        self.database.check()
        self.model.init()

    def init(self):
        log.info("Initializing %s", self)

        def state(cond):
            return cond and "enabled" or "disabled"

        log.info("Expiration is %s", state(self.expire_enabled))
        log.info("Vacuum is %s", state(self.vacuum_enabled))

        self.database.init()

        self.update_thread.init()

        # If self._qmf_classes has not been set, then build classes
        # here from the package list.  We want to be canonical, ie
        # always have things expressed in terms of classes
        if not self._classes:
            if not self._packages:
                self._packages = self.model._packages
            for pkg in self._packages:
                if pkg._name not in self._package_filter:
                    self._classes = self._classes.union(set(pkg._classes))
        else:            
            # Apply the package filter to the specified class list
            black_list = set()
            for cls in self._classes:
                if cls._package._name in self._package_filter:
                    black_list.add(cls)
            self._classes.difference_update(black_list)

        self.session.init(self._classes)

        if self.expire_enabled:
            self.expire_thread.init()

        if self.vacuum_enabled:
            self.vacuum_thread.init()

    def start(self):
        log.info("Starting %s", self)

        self.update_thread.start()

        self.session.start()

        if self.expire_enabled:
            # If we set these values here, we can have each cumin-report
            # instance run expiration for its own classes...
            # Default is like cumin-data, where a single instance runs
            # expiration across all classes in the model
            #self.expire_thread.packages = self.packages
            #self.expire_thread.classes = self.classes
            self.expire_thread.start()

        if self.vacuum_enabled:
            self.vacuum_thread.start()

    def stop(self, skip_vaccum=False):
        log.info("Stopping %s", self)

        self.update_thread.stop()
        self.session.stop()

        if self.expire_enabled:
            self.expire_thread.stop()

        if self.vacuum_enabled and not skip_vacuum:
            self.vacuum_thread.stop()

    def __repr__(self):
        return self.__class__.__name__
