import copy
import resource
import pickle

from psycopg2 import IntegrityError, TimestampFromTicks
from psycopg2.extensions import cursor as Cursor
from rosemary.model import *
from rosemary.sqltype import qmf_type_code_by_string, qmf_default_by_string

from model import *
from util import *

log = logging.getLogger("mint.update")

sample_window_min = 60
sample_window_max = 60 * 5

class UpdateThread(MintDaemonThread):
    def __init__(self, app):
        super(UpdateThread, self).__init__(app)

        self.updates = ConcurrentQueue(maxsize=1000)
        self.stats = UpdateStats(self.app)

        self.conn = None
        self.cursor = None

        self.halt_on_error = False

        # For timestamp persistent data, if this flag is True,
        # we will delete the objects and samples on startup.
        # This is to allow purging of large data sets.
        self.del_on_start = False

    def _delete_all_objects(self):
        def loop_body(cls):
            if cls._storage != "none":
                persistent = cls.check_persistent()
                if (persistent == "session") or \
                   (persistent == "timestamp" and self.del_on_start):
                    cls.delete_selection(self.cursor)
                    if persistent == "timestamp":
                        # del_on_start must be true...
                        cls.delete_sample_selection(self,cursor)
                    self.cursor.connection.commit()
                else:
                    log.debug("Skipping persistent class " + str(cls))

        classes = self.app.get_bound_classes()
        log.debug("Delete all objects by bound classes " + str(classes))
        for cls in classes:
            loop_body(cls)
 
    def init(self):
        self.conn = self.app.database.get_connection()

        self.cursor = self.conn.cursor(cursor_factory=UpdateCursor)
        self.cursor.stats = self.stats

    def enqueue(self, update):
        update.app = self.app
        self.updates.put(update)
        self.stats.enqueued += 1

    def run(self):
        self._delete_all_objects()

        while True:
            if self.stop_requested:
                break
            try:
                update = self.updates.get(True, 1)
            except Empty:
                continue
            if self.stop_requested:
                break

            self.stats.dequeued += 1

            update.process(self)

class UpdateStats(object):
    group_names = ("Updates", "Agents", "Objects")
    groups = "%43s | %32s | %32s |" % group_names

    heading_names = \
        ("Depth", "*Enqueued", "*Dequeued", "*Dropped",
         "*Created", "*Updated", "*Deleted",
         "*Created", "*Updated", "*Deleted",
         "*Sql Ops", "Errors", "Cpu (%)", "Mem (M)")
    headings_fmt = \
        "%10s %10s %10s %10s | " + \
        "%10s %10s %10s | " + \
        "%10s %10s %10s | " + \
        "%10s %10s %10s %10s"
    headings = headings_fmt % heading_names


    values_fmt = \
        "%10i %10.1f %10.1f %10.1f | " + \
        "%10.1f %10.1f %10.1f | " + \
        "%10.1f %10.1f %10.1f | " + \
        "%10.1f %10i %10i %10.1f"

    then = None
    now = None

    def __init__(self, app):
        self.enqueued = 0
        self.dequeued = 0
        self.dropped = 0

        self.agents_created = 0
        self.agents_updated = 0
        self.agents_deleted = 0

        self.objects_created = 0
        self.objects_updated = 0
        self.objects_deleted = 0

        self.objects_created_by_class = defaultdict(int)
        self.objects_updated_by_class = defaultdict(int)
        self.objects_deleted_by_class = defaultdict(int)

        self.sql_ops = 0
        self.errors = 0

        self.time = None
        self.cpu = 0
        self.memory = 0

    def capture(self):
        now = copy.copy(self)

        now.time = time.time()

        rusage = resource.getrusage(resource.RUSAGE_SELF) 

        now.cpu = rusage[0] + rusage[1]
        now.memory = self.get_resident_pages() * resource.getpagesize()

        UpdateStats.then = UpdateStats.now
        UpdateStats.now = now

    def get_resident_pages(self):
        try:
            line = open("/proc/%i/statm" % os.getpid()).read()
        except:
            return 0

        return int(line.split()[1])

    def print_headings(self):
        print self.groups
        print self.headings

    def print_values(self):
        self.capture()

        if not self.then:
            return

        values = [self.now.enqueued - self.then.enqueued,
                  self.now.dequeued - self.then.dequeued,
                  self.now.dropped - self.then.dropped,
                  self.now.agents_created - self.then.agents_created,
                  self.now.agents_updated - self.then.agents_updated,
                  self.now.agents_deleted - self.then.agents_deleted,
                  self.now.objects_created - self.then.objects_created,
                  self.now.objects_updated - self.then.objects_updated,
                  self.now.objects_deleted - self.then.objects_deleted,
                  self.now.sql_ops - self.then.sql_ops]

        # self.now.dropped - self.then.dropped,

        secs = self.now.time - self.then.time
        values = map(lambda x: x / secs, values)

        values.insert(0, self.now.enqueued - self.now.dequeued)

        values.append(self.errors)
        values.append(int((self.now.cpu - self.then.cpu) / secs * 100))
        values.append(self.now.memory / 1000000.0)

        print self.values_fmt % tuple(values)

    def print_values_by_class(self):
        names = ("Class", "Created", "Updated", "Deleted")
        print "%20s  %10s  %10s  %10s" % names

        for pkg in mint.model._packages:
            for cls in pkg._classes:
                created = stats.created_by_class[cls]
                updated = stats.updated_by_class[cls]
                deleted = stats.deleted_by_class[cls]

                if created or updated or deleted:
                    args = (cls._name, created, updated, deleted)
                    print "%-20s  %10i  %10i  %10i" % args
        
class UpdateCursor(Cursor):
    def execute(self, sql, args=None):
        super(UpdateCursor, self).execute(sql, args)

#        log.debug(self.mogrify(sql, args))

        self.stats.sql_ops += 1

class Update(object):
    def __init__(self, model):
        self.model = model
        self.app = None

    def process(self, thread):
        log.debug("Processing %s", self)

        try:
            self.do_process(thread.cursor, thread.stats)
            thread.conn.commit()
        except UpdateDropped:
            log.debug("Update dropped")

            thread.conn.rollback()

            thread.stats.dropped += 1
        except:
            log.debug("Update failed", exc_info=True)

            thread.conn.rollback()

            thread.stats.errors += 1

            #print_exc()

            if thread.halt_on_error:
                raise

    def do_process(self, cursor, stats):
        raise Exception("Not implemented")

    def __repr__(self):
        return self.__class__.__name__

class DeferredLink(object):
    def __init__(self, this, reference):
        self.this = this
        self.reference = reference

    def realize(self, cursor, that):
        column = self.reference.sql_column
        values = {column.name: that._id, "_id": self.this._id}

        self.this._class.sql_update_object.execute(cursor, values, (column,))

        msg = "Realized deferred link to %s via %s"
        log.debug(msg, that, self.reference)

class AviaryUpdate(Update):
    def __init__(self, model, aviary_obj, cls):
        super(AviaryUpdate, self).__init__(model)

        self.aviary_obj = aviary_obj
        self.cls = cls
        self.delete = False

    def do_process(self, cursor, stats):
        agent_id = self.get_agent_id()
        object_id = self.get_object_id()

        # For aviary stuff, the agent is inferred
        # There really are no separate updates, so
        # we just base the agent id on info in the object
        try:
            agent = self.model.agents_by_id[agent_id]
        except KeyError:
            if self.delete:
                raise UpdateDropped()
            agent = MintAgent(self.model, agent_id)
            stats.agents_created += 1
        
        try:
            obj = agent.get_object(cursor, self.cls, object_id)
        except RosemaryNotFound:
            if self.delete:
                raise UpdateDropped()
            obj = self.create_object(cursor, stats, self.cls)
            return

        if self.delete:
            self.delete_object(cursor, stats, obj)
            del agent.objects_by_id[obj._qmf_object_id]
        else:
            self.update_object(cursor, stats, obj)

    def get_agent_id(self):
        return str(self.aviary_obj.id.resource) + ":" + \
               str(self.aviary_obj.id.name)

    def get_object_id(self):
        return str(self.aviary_obj.id.name)

    def delete_object(self, cursor, stats, obj):
        obj.delete(cursor)
        self.model.print_event(3, "Deleted %s", obj)
        stats.objects_deleted += 1

    def create_object(self, cursor, stats, cls):
        update_time = datetime.now()
        create_time = update_time

        obj = cls.create_object(cursor)
        obj._qmf_agent_id = self.get_agent_id()
        obj._qmf_object_id = self.get_object_id()
        obj._qmf_update_time = update_time

        object_columns = list()
        sample_columns = list()

        table = cls.sql_table

        object_columns.append(table._id)
        object_columns.append(table._qmf_agent_id)
        object_columns.append(table._qmf_object_id)
        object_columns.append(table._qmf_update_time)

        self.process_properties(obj, object_columns, cursor)
        self.process_statistics(obj, object_columns, sample_columns)

        statements = list()

        sql = cls.sql_insert_object.emit(object_columns)
        statements.append(sql)

        if sample_columns:
            sample_columns.append(cls.sql_samples_table._qmf_update_time)

            sql = cls.sql_samples_insert.emit(sample_columns)
            statements.append(sql)

            obj._sample_time = time.time()

        sql = "; ".join(statements)
        self.execute_sql(cursor, sql, obj.__dict__)

        self.process_deferred_links(cursor, obj)

        obj._save_time = datetime.now()

        self.model.print_event(3, "Created %s", obj)

        stats.objects_created += 1
        #stats.objects_created_by_class[cls] += 1

        return obj

    def update_object(self, cursor, stats, obj):
        update_time = datetime.now()
        create_time = update_time

        obj._qmf_update_time = update_time

        object_columns = list()
        sample_columns = list()

        self.process_properties(obj, object_columns, cursor)
        self.process_statistics(obj, object_columns, sample_columns)

        statements = list()
        cls = obj._class

         # force a write if it's been too long, even if the values match
        if object_columns \
                or (obj._save_time != None and \
                    obj._qmf_update_time != None and \
                    obj._save_time < obj._qmf_update_time - timedelta(hours=1)):
            object_columns.append(cls.sql_table._qmf_update_time)

            sql = cls.sql_update_object.emit(object_columns)
            statements.append(sql)

            obj._save_time = datetime.now()

        if sample_columns:
            sample_columns.append(cls.sql_samples_table._qmf_update_time)

            sql = cls.sql_samples_insert.emit(sample_columns)
            statements.append(sql)

            obj._sample_time = time.time()

        if not statements:
            raise UpdateDropped()

        sql = "; ".join(statements)
        self.execute_sql(cursor, sql, obj.__dict__)

        self.model.print_event(4, "Updated %s", obj)

        stats.objects_updated += 1
        #stats.objects_updated_by_class[cls] += 1

    def transform_value(self, prop, value):
        # For some reason the timestamps in QMF values
        # are in 10s of microseconds, so we need to
        # handle our own conversion here
        if prop.type == "absTime":
            return datetime.fromtimestamp(float(value))
        t = qmf_type_code_by_string[prop.type]
        return transformers[t](value)

    def get_default(self, prop):
        if prop.type in qmf_default_by_string:
            return qmf_default_by_string[prop.type]
        return None

    def process_deferred_links(self, cursor, obj):
        agent = self.model.agents_by_id[obj._qmf_agent_id]

        if obj._qmf_object_id not in agent.deferred_links_by_id:
            return

        links = agent.deferred_links_by_id[obj._qmf_object_id]

        for link in links:
            link.realize(cursor, obj)

        del agent.deferred_links_by_id[obj._qmf_object_id]


    def process_properties(self, obj, columns, cursor):
        cls = obj._class
   
        for name, prop in cls._properties_by_name.iteritems():
            try:
                value = getattr(self.aviary_obj, name)
            except AttributeError:
                value = self.get_default(prop)

            if value is not None:
                value = self.transform_value(prop, value)
            if value == getattr(obj, prop.sql_column.name):
                continue
            setattr(obj, prop.sql_column.name, value)
            columns.append(prop.sql_column)

    def process_reference(self, obj, prop, oid, cursor):
        try:
            ref = obj._class._references_by_name[prop.name]
        except KeyError:
            raise MappingException("Reference %s is unknown" % prop.name)

        if not ref.sql_column:
            raise MappingException("Reference %s has no column" % ref.name)

        # Just create the agent if we haven't seen it.
        value = None
        try:
            agent = self.model.agents_by_id[oid.agentName]
        except KeyError:
            agent = MintAgent(self.model, oid.agentName)

        object_id = oid.objectName
        try:
            that = agent.get_object(cursor, ref.that_cls, object_id)
        except RosemaryNotFound:
            link = DeferredLink(obj, ref)
            agent.deferred_links_by_id[object_id].append(link)

            msg = "Deferring link to object %s %s"
            raise MappingException(msg % (ref.that_cls, object_id))

        value = that._id

        return ref.sql_column, value

    def process_statistics(self, obj, update_columns, insert_columns):
        build_columns = list()
        saw_change = False

        # get a list of statistic names from the cls
        # loop through the list of properties, look for
        # the value in the object
        for name, stat in self.cls._statistics_by_name.iteritems():
            try:
                value = getattr(self.aviary_obj, name)
            except AttributeError:
               value = self.get_default(stat)

            if value is not None:
                value = self.transform_value(stat, value)

            # Don't write unchanged values
            if value != getattr(obj, stat.sql_column.name):
                setattr(obj, stat.sql_column.name, value)
                update_columns.append(stat.sql_column)
                saw_change = True

            # If we do end up seeing a value change, we will
            # need to insert an entire row so build it up
            build_columns.append(stat.sql_column)

        if saw_change:        
            insert_columns.extend(build_columns)

    def execute_sql(self, cursor, text, args):
        try:
            cursor.execute(text, args)
        except:
            log.debug("%s failed sql execute", self, exc_info=True)
            log.error("Sql execute failed")
            log.error("Sql text: %s", text)
            log.error("Sql values:")

            for item in sorted(args.items()):
                log.error("    %-34s  %r", *item)

            log.error("Sql row count: %i", cursor.rowcount)

            log.error("attributes:")

            for item in sorted(self.aviary_obj.__dict__.items()):
                log.error("    %-34s  %r" % item)
            raise

    def __repr__(self):
        name = self.__class__.__name__
        agent_id = self.get_agent_id()
        obj_id = self.get_object_id()

        return "%s(%s,%s,%s)" % (name, agent_id, self.cls, obj_id)

class AviaryScheddReferencer(AviaryUpdate):
    def __init__(self, model, aviary_obj, cls):
        super(AviaryScheddReferencer, self).__init__(model, aviary_obj, cls)

    def process_properties(self, obj, columns, cursor):
        cls = obj._class
   
        # Fix up submissions to hold a scheduler ref
        # remember that we saw this
        ref = cls._references_by_name["schedulerRef"]

        class FakeProp(object):
            def __init__(self, name):
                self.name = name

        class FakeOID(object):
            def __init__(self, agent, objname):
                self.agentName = agent
                self.objectName = objname

        # Okay, we *know* how a scheduler is going to name itself
        # so we invent a reference here so that it can be resolved
        # using current machinery
        sched_name = self.get_scheduler_name()
        val = FakeOID("SCHEDULER:" + sched_name, sched_name)

        prop = FakeProp("schedulerRef")
        try:
            col, nvalue = self.process_reference(obj, prop, val, cursor)
            if nvalue != getattr(obj, col.name):
                setattr(obj, col.name, nvalue)
                columns.append(col)
        except MappingException, e:
            log.debug(e)

        super(AviaryScheddReferencer, self).process_properties(obj, columns, cursor)

    def get_agent_id(self):
        return "SCHEDULER" + ":" + \
               self.get_scheduler_name()

    def get_scheduler_name(self):
        raise Exception("NotImplemented")

class AviarySubmissionUpdate(AviaryScheddReferencer):
    def __init__(self, model, submission):
        super(AviarySubmissionUpdate, self).__init__(model, 
                                                     submission,  
                                                     model.com_redhat_grid.Submission)

    def get_object_id(self):
        return str(self.aviary_obj.id.owner) + ":" + \
               str(self.aviary_obj.id.name)

    def get_scheduler_name(self):
        return str(self.aviary_obj.id._scheduler)


class AviaryCollectorUpdate(AviaryUpdate):
    def __init__(self, model, aviary_obj):

        pkg = model._packages_by_name[aviary_obj.pkg_name]
        cls = pkg._classes_by_lowercase_name[aviary_obj.class_name.lower()]
        

        super(AviaryCollectorUpdate, self).__init__(model, 
                                                    aviary_obj,  
                                                    cls)

class AviarySubmitterUpdate(AviaryScheddReferencer):
    #  This is very much like a AviaryCollectorUpdate but it needs to have
    # a schedd reference injected so we derive from AviaryScheddReferencer
    def __init__(self, model, aviary_obj):
        
        pkg = model._packages_by_name[aviary_obj.pkg_name]
        cls = pkg._classes_by_lowercase_name[aviary_obj.class_name.lower()]

        super(AviarySubmitterUpdate, self).__init__(model, aviary_obj, cls)

    def get_object_id(self):
        return str(self.aviary_obj.id.name)

    def get_scheduler_name(self):
        return str(self.aviary_obj.id.scheduler)

class UpdateDropped(Exception):
    pass

class MappingException(Exception):
    pass

def transform_default(value):
    return value

def transform_timestamp(value):
    if value != 0:
        return datetime.fromtimestamp(value / 1000000000)

def transform_pickle(value):
    return pickle.dumps(x)

transformers = list([transform_default for x in range(32)])

transformers[8] = transform_timestamp
transformers[10] = str
transformers[14] = str
transformers[15] = transform_pickle

def transform_value(attr, value):    
    return transformers[attr.type](value)
