import logging
import time
import copy
from Queue import Queue, Empty

from sage.util import CallThread, parse_URL
from threading import Lock, Condition

#import cProfile
import wallaroo

log = logging.getLogger("sage.wallaby")

class WBTypes:
    '''
    Wrap symbolic names for data items tracked by WallabyOperations.
    May be helpful in avoiding typos, good for interpreter help
    '''
    NODES = "nodes"
    GROUPS = "groups"
    TAGS = "tags"
    FEATURES = "features"

class WallabyOperations(object):
    '''
    Wrapper around the Wallaby client library.
    '''
    def __init__(self, wallaroo_uri, refresh_interval=None):
        '''
        Constructor.

        wallaroo_uri -- the URI used to connect to a wallaroo server.
        The simplest URI is just a hostname but a full URI can specify 
        scheme://user/password@host:port or a subset of those components 
        as long as the host is included.

        Examples:

            localhost
            localhost:5672
            http://fred/flintsone@quarry.bedrock.com:1234

        refresh_interval -- maximum time WallabyOperations will sleep
        before checking that if the store has changed outside of cumin.
        Any time the store is modified in cumin, the check will be made
        '''
        self.wallaroo_uri = wallaroo_uri

        # The cache maintenance thread.  Most of the attributes here
        # are managed by the cache maintenance thread.
        self._maintain_cache = None
        self.refresh_interval = refresh_interval

        self._store = None
        self._wallaroo_tag = None

        # The replay list for adding unactivated changes to current after
        # an update.  An activation command will clear this list.
        self._replay_list = []
        self._replaying = False

        # Can be set by an individual command to keep it off the replay list
        self._skip_replay = False

        # the command queue to the maintenance thread
        self.queue = Queue()

        # Stop the maintenance thread
        self._stop = False

        self._cache = Cache()

        # Store the name of the partition group so we can filter it out
        # of tags/groups that we return
        self._partition_group = None

        # Lock is used for synchronization with the caching thread and
        # for thread safety of any and all data that could be accessed
        # by multiple threads.
        self._lock = Lock()
        self._condition = Condition(self._lock)

    # This is a decorator to cause operations to be queued    
    def _queued(op):
        def wrap(self, *args, **kwargs):
            def enc():
                op(self, *args, **kwargs)
            self.queue.put(enc)
        return wrap

    def start(self, retry_secs=5):
        '''
        Start the caching thread.

        This thread will attempt to connect to wallaroo and retrieve
        a Store object.  If successful, it will periodically retrieve 
        and cache data from wallaroo.

        Only one caching thread may run at a time.  The thread may
        be restarted if it has previously been stopped.

        Note, for the moment start() and stop() are not thread safe.  They
        should only be called from a single thread.

        retry_secs -- how often the caching thread will retry failed
        operations.  This includes attempts to connect to wallaroo
        and retrieve a Store object as well as calls to wallaroo that
        return no data.
        '''            
        # Wrap the entire cache thread with an exception handler
        def wrap_maintain_cache(retry_secs):
           try:
              self.maintain_cache(retry_secs)
              log.debug("WallabyOperations: cache maintenance thread exited")
           except Exception:
              import traceback
              traceback.print_exc()

        if self._maintain_cache is not None and \
           self._maintain_cache.isAlive():
            # No, you can't start another one.
            return False

        self._stop = False

        if self.wallaroo_uri is not None:
           self._maintain_cache = CallThread(wrap_maintain_cache, None, retry_secs)
           self._maintain_cache.daemon = True
           self._maintain_cache.start()
           log.debug("WallabyOperations: start cache maintenance thread")
           return True
        return False

    def stop(self, wait=False, timeout=None):
        '''
        Stop the caching thread.

        Wake the caching thread if asleep and cause it to exit.
        The thread may be restarted again with a call to start()
        once it has successfully exited.  On exit, the thread will
        null out cached data.

        wait -- if True the call will block until the thread exits or
        "timeout" seconds has passed if "timeout" is not None.

        timeout -- how long to wait for the thread to exit if "wait" is True.
        A value of None means wait forever.

        Note, for the moment start() and stop() are not thread safe.  They
        should only be called from a single thread.
        '''
        if self._maintain_cache is not None:
            self._stop = True

            # could be sleeping on the queue, wake it up
            self.queue.put(True) 

            # could be sleeping on the condition
            self._condition.acquire()
            self._condition.notify()
            self._condition.release()

            if wait and self._maintain_cache.isAlive():
                log.debug("WallabyOperations: waiting for cache maintenance thread to exit")
                self._maintain_cache.join(timeout)
            log.debug("WallabyOperations: stopped cache maintenance thread")

    def get_data(self, which, valuefilter=None):
        '''
        Return a list of cached values for the specified category.

        The values returned will be proxy objects constructed by
        the Wallaby client library.

        which -- specifies the category.  Attributes of WBTypes 
        define valid values for "which"
        '''
        d = []
        self._lock.acquire()
        try:
            if which in self._cache.data:
                d = self._cache.data[which].values()
            # Here we handle the possible filtering of node names
            if which == WBTypes.NODES:
                if valuefilter is not None and valuefilter["nodeName"] != "%%%":
                    filter = valuefilter["nodeName"].replace("%", "")
                    if filter != "":
                        d = [value for value in d if value.name.find(filter) > -1]
        finally:
            self._lock.release()
        return d

    def get_names(self, which):
        '''
        Return a list of cached names for the specified category.

        The values returned will be the names of objects constructed
        by the Wallaby client library.

        which -- specifies the category.  Attributes of WBTypes 
        define valid values for "which"
        '''
        d = []
        self._lock.acquire()
        try:
            if which in self._cache.data:
                d = self._cache.data[which].keys()
        finally:
            self._lock.release()
        return d            

    def get_node_by_name(self, name):
        '''
        Return a cached wallaby.Node object by name.

        If name does not designate a currently cached
        object, None is returned.
        '''
        return self._lookup_by_name(WBTypes.NODES, name)

    def get_group_by_name(self, name):
        '''
        Return a cached wallaby.Group object by name.

        If name does not designate a currently cached
        object, None is returned.
        '''
        return self._lookup_by_name(WBTypes.GROUPS, name)

    def get_tag_by_name(self, name):
        '''
        Return a cached wallaby.Tag object by name.

        If name does not designate a currently cached
        object, None is returned.
        '''
        return self._lookup_by_name(WBTypes.TAGS, name)

    def get_feature_by_name(self, name):
        '''
        Return a cached wallaby.Feature object by name.

        If name does not designate a currently cached
        object, None is returned.
        '''
        return self._lookup_by_name(WBTypes.FEATURES, name)

    def get_node_names(self, tag_name):
        '''
        Return a list of node names associated with the tag.

        The return result is a list containing the names of nodes
        in the tag group.
        '''
        names = []
        self._lock.acquire()
        try:
            if tag_name in self._cache._nodes_by_tag:
                names =  copy.copy(self._cache._nodes_by_tag[tag_name])
        finally:
            self._lock.release()
        return names

    def get_tag_names(self, node_name):
        '''
        Return a list of tag names associated with the node.

        The return result is a list containing the names of tags
        on the specified node.
        '''
        names = []
        self._lock.acquire()
        try:
            if node_name in self._cache._tags_by_node:
                names = copy.copy(self._cache._tags_by_node[node_name])
        finally:
            self._lock.release()
        return names

    @_queued
    def create_tags(self, names, callback=None):
        '''
        Create new tags in the Wallaby store.

        Refresh the cached lists of groups and tags.
        '''
        tags = []
        try:                    
            for name in names:
                start = time.time()
                self._store.addLabel(name)
                tags.append(self._store.getGroupByName(name))
                delta = time.time() - start
                log.debug("WallabyOperations: store.addLabel %s" % delta)
        except Exception, e:
            log.debug("WallabyOperations: create_tag, "\
                      "exception suppressed, %s" % str(e))

        # If we actually created tags then we must add them to the cache
        # Each tag must go in the groups and tags cache, and the tag has
        # no nodes yet because we just made it.

        # any @_queued method is a replayable method
        # we have to select the correct cache and lock conditionally
        cache = self._queued_cache()
        self._acquire_if()
        for tag in tags:
            cache._nodes_by_tag[tag.name] = []
            cache.data["tags"][tag.name] = tag
            cache.data["groups"][tag.name] = tag
        self._release_if()

        if callback and not self._replaying:
            callback(len(tags) > 0)

    @_queued
    def remove_tags(self, names, callback=None):
        '''
        Remove a set of tags from the Wallaby store.

        Check the cached list of tags for the
        tag name first.  Refresh cached lists of
        groups, tags, and nodes.
        '''

        # any @_queued method is a replayable method
        # we have to select the correct cache and lock conditionally
        cache = self._queued_cache()
        tags = []
        for name in names:
            if name in cache.data["tags"]:
                try:
                    # This handler here is a hack to allow remove group
                    # when it really doesn't work yet
                    try:
                        self._store.removeGroup(name)
                    except:
                        pass
                    tags.append(name)
                except Exception, e:
                    log.debug("WallabyOperations: remove_tag, "\
                              "exception suppressed, %s" % str(e))

        # If we removed tags then we must remove them from the cache
        self._acquire_if()
        for tag in tags:
            #for n in self._cache._nodes_by_tag[tag]:
            #    n.refresh()
            nodes = cache._nodes_by_tag[tag]
            del cache._nodes_by_tag[tag]
            del cache.data["tags"][tag]
            del cache.data["groups"][tag]
            for n in nodes:
                try:
                    cache._tags_by_node[n].remove(tag)
                except:
                    pass
        self._release_if()

        if callback and not self._replaying:
            callback(len(tags) > 0)

    @_queued
    def edit_tags(self, node, *tags, **kwargs):
        '''
        Replace existing tags on a node with the specified tags.

        node -- a wallaby.Node object or the name of a wallaby.Node object

        tags -- the new set of tags for the node, 
                list or tuple of strings
        '''

        # any @_queued method is a replayable method
        # we have to select the correct cache and lock conditionally
        cache = self._queued_cache()
        status = False
        n = node in cache.data[WBTypes.NODES] and \
            cache.data[WBTypes.NODES][node] or None
        if n is None:
            log.debug("WallabyOperations: edit_tags, node not found %s" % str(n))
        else:
            try:
                # cast to list in case tags is a tuple, list required
                start = time.time()
                # this is a workaround
                self._store.getPartitionGroup() 

                # In case multiple cumin users are fighting
                # check the list of tags for validity
                bogus = []
                for t in tags:
                    if not t in cache.data[WBTypes.TAGS]:
                        bogus.append(t)
                for b in bogus:
                    tags.remove(b)

                n.modifyLabels("REPLACE", list(tags))
                status = True
                delta = time.time() - start
                log.debug("WallabyOperations: edit_tags %s" % delta)
            except Exception, e:
                log.debug("WallabyOperations: edit_tags, "\
                              "exception suppressed, %s" % str(e))

            # We know before we start which tags were on the node (tags_by_node[node])                
            # We know the tags we just set (tags)
            if status:
                orig = node in cache._tags_by_node and \
                       set(cache._tags_by_node[node]) or set([])
                new = set(tags)

                added_tags = new - orig
                removed_tags = orig - new

                self._acquire_if()
                cache._tags_by_node[node] = list(tags)

                for a in added_tags:
                    cache._nodes_by_tag[a].append(node)

                for r in removed_tags:
                    cache._nodes_by_tag[r].remove(node)
                self._release_if()

        callback = "callback" in kwargs and kwargs["callback"] or None
        if callback and not self._replaying:
            callback(status)

    @_queued
    def edit_features(self, group, *features, **kwargs):
        '''
        Replace existing features in a group with the specified features
        group -- a wallaby.Group object or the name of a wallaby.Group object

        features -- the new set of features for the group, 
                    list or tuple of strings
        '''

        # any @_queued method is a replayable method
        # we have to select the correct cache and lock conditionally
        cache = self._queued_cache()
        status = False
        g = group in cache.data[WBTypes.GROUPS] and \
            cache.data[WBTypes.GROUPS][group] or None
        if g is None:
            log.debug("WallabyOperations: edit_features, group not found %s" % str(g))
        else:
            try:
                start = time.time()
                g.modifyFeatures("REPLACE", features, {})
                status = True
                delta = time.time() - start
                log.debug("WallabyOperations: edit_features %s" % delta)
            except Exception, e:
                log.debug("WallabyOperations: edit_features, exception suppressed, %s" % str(e))

        # no other data to update here because we don't have a mapping
        # of features to groups in the cache
        callback = "callback" in kwargs and kwargs["callback"] or None
        if callback and not self._replaying:
            callback(status)

    @_queued
    def activate_configuration(self, callback=None):
        '''
        Call activateConfiguration on the store object.
        Returns False on failure otherwise True.
        '''

        # Do not allow this command to be added to the replay
        self._skip_replay = True

        try:
            if self._my_branch_is_current():
                status = True
            else:
                status = self._store.activateConfiguration()
                # If wallaroo gives us something that evaluates to False fix it
                if not status:
                    status = True
        except:
            status = False

        if status:
            # Make sure our branch points at the current tag
            self._update_branch_to_current()

            # Remember the new wallaroo current tag
            self._wallaroo_tag = self._get_current_wallaroo_tag()

            # Clear the replay list, because we activated!
            self._replay_list = []

        if callback:
            callback(status)

    # Super secret private implementation stuff.  Don't look!

    def _get_current_wallaroo_tag(self):
       return self._store.cm.fetch_json_resource("/tags/current")

    def _compare_tags(self, first, second):
        try:
            return first["commit"] == second["commit"]
        except:
            pass
        return False

    def _my_branch_is_current(self):
       my_branch = self._store.cm.how.to_q()
       return 'tag' in my_branch and my_branch['tag'] == 'current'

    def _update_branch_to_current(self):
        self._store.cm.how = wallaroo.client.cmeta.mk_how({"tag": "current"})

    def _connect(self, retry_secs, connect_args):
        # Get connection and store obect
        while not self._stop:
            log.debug("WallabyOperations: attempting to get Wallaby store object")
            try:
               self._store = wallaroo.client.connect(**connect_args)
            except:
               pass
            if self._store is not None:
               self._partition_group = self._store.getPartitionGroup().name
               log.debug("WallabyOperations: found wallaby store object")
               break

            # Check stop inside the lock to make sure that we don't miss
            # a signal or a "stop" that was set while we were iterating.
            self._condition.acquire()
            if not self._stop:
               self._condition.wait(retry_secs)
            self._condition.release()

    def _queued_cache(self):
        if self._replaying:
            return self._replay_cache
        else:
            return self._cache

    def _acquire_if(self):
        if not self._replaying:
            self._lock.acquire()

    def _release_if(self):
        if not self._replaying:
            self._lock.release()

    def _recache(self, items=[]):

        # Make a new cache to hold updates
        cache = Cache()

        # Mark explict refresh
        for attr in cache.data:
            if not items or attr in items:
                cache.set_refresh(attr, True)

        cache.update(self._store)

        return cache

    def _replay_pending(self, cache):
        if not self._replay_list:
            return

        self._replaying = True
        self._replay_cache = cache
        for job in self._replay_list:
            try:
                job()
            except:
                pass
        self._replaying = False
        self._replay_list = []
        self._replay_cache = None

    def _make_current(self):
        while True:
            wallaroo_current_tag = self._get_current_wallaroo_tag()
            if self._compare_tags(wallaroo_current_tag, 
                                  self._wallaroo_tag):
                break
            else:
                self._wallaroo_tag = wallaroo_current_tag

                # Change to the current tag if we are working on a local branch.  
                # If we are already current then we can skip this step.
                if not self._my_branch_is_current():
                    self._update_branch_to_current()
                    cache = self._recache()
                    self._replay_pending(cache)
                else:
                    # We were looking at current, but there was an activation
                    # so we don't know everything.  We can just recache since
                    # we don't have any unactivated changes
                    cache = self._recache()

                # Now that we're all up to date swap in the new cache atomically
                self._lock.acquire()
                self._cache.swap(cache)
                self._lock.release()

    def maintain_cache(self, retry_secs):

        # Turn the URL into a dictionary of args that the wallaroo
        # client.connect() method wants to see
        url = parse_URL(self.wallaroo_uri)
        arg_map = (("host",),
                   ("port",),
                   ("scheme",),
                   ("user",     "username"),
                   ("password", "pw"))

        connect_args = dict()
        for arg in arg_map:
            v = getattr(url, arg[0])
            if v is not None:
                connect_args[arg[-1]] = v

        # if we have an error, we should try a reconnect maybe
        while not self._stop:
            self._connect(retry_secs, connect_args)
            try:
                self._make_current()
                while not self._stop:
                    try:
                        job = self.queue.get(timeout=self.refresh_interval)
                    except Empty:
                        job = None
                    if self._stop:
                        break

                    # Update the store object if it has changed outside of 
                    # cumin and then recache all of the objects
                    self._make_current()

                    # Now that we know everything is up to date do the operation
                    # and store it on the replay list
                    if job:
                        try:
                            job()
                        except:
                            pass

                        # The job may have set skip replay
                        if self._skip_replay:
                            self._skip_replay = False
                        else:
                            self._replay_list.append(job)

            except Exception, e:
                # The point here is to allow a reconnect
                # to the wallaby store.  We may want to filter on particular
                # exception types.  Hmmm, what do we reinitialize? raise for now
                self._lock.acquire()
                self._cache.clear()
                self._lock.release()
                log.debug("WallabyOperations: Reconnecting, exception is %s" % str(e))

        # Clear cache if we have been stopped....
        self._lock.acquire()
        self._cache.clear()
        self._store = None
        self._lock.release()

    def _lookup_by_name(self, which, name):
        n = None
        self._lock.acquire()
        try:
            if name in self._cache.data[which]:
                n = self._cache.data[which][name]
        finally:
            self._lock.release()
        return n

class CacheData(object):
    def __init__(self, synthetic=False, args=()):
        # These attributes are only referenced from the cache thread
        # so they do not need to be protected
        self.synthetic = synthetic
        self.refresh = False

class NodeData(CacheData):
   def get_values(self, store, cache):
       start = time.time()
       nodes = [(name, store.getNode(name)) for name in store.nodes]
       delta = time.time() - start
       log.debug("WallabyOperations: get node names and nodes %s" % delta)
       for i in nodes:
           i[1].last_checkin = time.time()
       cache.data["nodes"] = dict(nodes)

class GroupData(CacheData):
   def get_values(self, store, cache):
      start = time.time()
      stuff = [(name, store.getGroupByName(name)) for name in store.groups]
      delta = time.time() - start
      log.debug("WallabyOperations: get group names and groups %s " % delta)
      cache.data["groups"] = dict(stuff)

class FeatureData(CacheData):
   def get_values(self, store, cache):
       start = time.time()
       stuff = [(name, store.getFeature(name)) for name in store.features]
       delta = time.time() - start
       log.debug("WallabyOperations: get feature names and features %s " % delta)               
       cache.data["features"] = dict(stuff)

class TagData(CacheData):
   def get_values(self, store, cache):
      # figure out the tag list and nodes per tag
      groups = cache.data["groups"].values()
      tags = []
      nodes_by_tag = dict()
      tags_by_node = dict()
      for g in groups:
          if not g.name.startswith("+++") and \
             store.isLabel(g):
              tags.append((g.name, g))
              start = time.time()
              nodes = g.membership
              delta = time.time() - start
              log.debug("WallabyOperations: update g.membership %s " % delta)
              nodes_by_tag[g.name] = nodes
              for n in nodes:
                  if n in tags_by_node:
                      tags_by_node[n].append(g.name)
                  else:
                      tags_by_node[n] = [g.name]

      cache.data["tags"] = dict(tags)
      cache._nodes_by_tag = nodes_by_tag
      cache._tags_by_node = tags_by_node

      log.debug("WallabyOperations: %s list updated (%s items)" \
                   % ("tags", len(tags)))

class Cache(object):

    gen = {WBTypes.NODES:    NodeData(),
           WBTypes.GROUPS:   GroupData(),
           WBTypes.FEATURES: FeatureData(),
           WBTypes.TAGS:     TagData(synthetic=True)}

    def __init__(self):
        self._nodes_by_tag = dict()
        self._tags_by_node = dict()

        self.data = {WBTypes.NODES:    dict(),
                     WBTypes.GROUPS:   dict(),
                     WBTypes.FEATURES: dict(),
                     WBTypes.TAGS:     dict()}

    def update(self, store):                
        for attr, val in self.gen.iteritems():
            if val.refresh and not val.synthetic:
                val.get_values(store, self)

        for attr, val in self.gen.iteritems():
            if val.refresh and val.synthetic:
                val.get_values(store, self)

    def swap(self, cache):
        # These lists are synthesized from tag data
        # Copy the values if tags was refreshed
        if self.gen["tags"].refresh:
            self._nodes_by_tag = cache._nodes_by_tag
            self._tags_by_node = cache._tags_by_node

        for attr, val in self.gen.iteritems():
            if val.refresh:
                self.data[attr] = cache.data[attr]
            val.refresh = False

    def set_refresh(self, attr, flag):
        self.gen[attr].refresh = flag

    def clear(self):
        for attr in self.data:
            self.data[attr] = dict()
